#include <opencv2/opencv.hpp>
#include <opencv2/ml.hpp>
#include <opencv2/xfeatures2d.hpp>
#include <opencv2/features2d/features2d.hpp>


#include <iostream>
#include <fstream>

#include "Svm.h"


using namespace std;
using namespace cv;
using namespace cv::ml;


#define num_train_sample  200
#define num_test_sample   200


#define with_steps

#define features_size 128  //para sift    

#define tam_dicionario 100

void readLabels(Mat &train_l, Mat &test_l);
Mat getVocabulary(Feature2D *detector);
void getDescritionBOW(
            Mat vocabulary,  //vocabulario gerado por getVocabulary
            String matcher,  //tipo do matcher ()
            Feature2D *detector, //detector de pontos (SIFT, SURF, ...)
            Mat &matTrainFeatures, //resultado para treinamento
            Mat &matTestFeatures); //resultado para teste


#define with_steps

int main() {

    //features
    Mat matTrainFeatures(0,0,CV_32F); //alocado dinamicamente
    Mat matTestFeatures(0,0,CV_32F);
    
    //labels
    Mat matTrainLabels= Mat::zeros(1,num_train_sample,CV_32SC1);
    Mat matTestLabels= Mat::zeros(1,num_test_sample,CV_32SC1);

#ifdef with_steps
    cout << "Lendo labels ..." << endl;
#endif
    //obtem labels do mnist
    readLabels(matTrainLabels,matTestLabels);
    
    //instancia detector, aqui sift. Poderia ser qualquer coisa
    Ptr<Feature2D> detector = xfeatures2d::SIFT::create();
    
    //gera vocabulario
    //atente que se pode optar por salvar o vocabulario 
    //e executar vários testes

#ifdef with_steps
    cout << "Gerando vocabulário ..." << endl;
#endif
    Mat vocabulary = getVocabulary(detector);
    //TODO você poderia salvar a variável vocabulary em um arquivo
    
    
#ifdef with_steps
    cout << "Carregando features ..." << endl;
#endif
    //carrega features usando vocabulario gerado
    getDescritionBOW(
            vocabulary,  //vocabulario gerado por getVocabulary
            "FlannBased",  //tipo do matcher (BruteForce, BruteForce-L1, BruteForce-Hamming, FlannBased)
            detector, //detector de pontos (SIFT, SURF, ...)
            matTrainFeatures, //resultado para treinamento
            matTestFeatures);
    
    //instacia e usa svm
    Svm svm;
    
#ifdef with_steps
    cout << "Treinando SVM ..." << endl;
#endif
    svm.loadTrainAndAdjust(matTrainFeatures, matTrainLabels);
    svm.predict(matTestFeatures);

#ifdef with_steps
    cout << "Gerando Resultados ..." << endl;
#endif
    svm.checkResults(matTestLabels, 10);
    
    return 0;
}


Mat getVocabulary(Feature2D *detector) {
    
    int i;
    vector<KeyPoint> keypoints; 
    Mat descriptors;
    ostringstream nome;
    
    Mat train=Mat::zeros(0,0, CV_32F);
    
    for (i=1; i<=num_train_sample; i++){
        nome.str("");
        nome << "train/" << i << ".png";
        Mat im = imread(nome.str(),0);
        
        //obtem features
        detector->detectAndCompute( im,  Mat(), keypoints, descriptors );
        train.push_back(descriptors);
    }
    
    //instancia BOW
    BOWKMeansTrainer bowtrainer(tam_dicionario); //num clusters
    //adiciona amostra de treinamento para gerar dicionario
    bowtrainer.add(train);
    //cria dicionario
    Mat vocabulary = bowtrainer.cluster();
    return vocabulary;
}


void getDescritionBOW(Mat vocabulary, 
            String tipoMatcher, 
            Feature2D *detector, 
            Mat &matTrainFeatures,
            Mat &matTestFeatures) {
    
    int i;
    vector<KeyPoint> keypoints; 
    Mat descriptors;
    ostringstream nome;
    
    //guarda o vetor de características
    //histograma que representa a quantidade de matches
    Mat response_hist; 
    
    //instancia o matcher que fará a correspondencia do vocabulário
    //com as features da imagem
    Ptr<DescriptorMatcher> matcher = DescriptorMatcher::create(tipoMatcher);
    
    //define o extrato padrão a ser utilizado
    Ptr<DescriptorExtractor> extractor(detector);
    
    //Instacia o descritor do BOW
    BOWImgDescriptorExtractor bowide(extractor,matcher);
    bowide.setVocabulary(vocabulary); //ajusta vocabulário
 
    for (i=1; i<=num_train_sample; i++){
        nome.str("");
        nome << "train/" << i << ".png";
        Mat im = imread(nome.str(),0);
        
        detector->detectAndCompute( im,  Mat(), keypoints, descriptors );

        //descreve as features de acordo com o dicionario
        bowide.compute(im, keypoints,response_hist);
        if (response_hist.rows == 0)  //caso não contenha features
            response_hist= Mat::zeros(1,tam_dicionario,CV_32F);
        matTrainFeatures.push_back(response_hist);
    }
    
    for (i=6001; i<=6000+num_test_sample; i++){
        nome.str("");
        nome << "test/" << i << ".png";
        Mat im = imread(nome.str(),0);
        
        
        detector->detectAndCompute( im,  Mat(), keypoints, descriptors );
        
        //descreve as features de acordo com o dicionario
        bowide.compute(im, keypoints, response_hist);
        if (response_hist.rows == 0) //caso não contenha features
            response_hist= Mat::zeros(1,tam_dicionario,CV_32F);
        
        matTestFeatures.push_back(response_hist);
    }
}

void readLabels(Mat &train_l, Mat &test_l) {
    
    ifstream labels ("labels.txt");
    int elm;
    if (labels.is_open()) 
        
    
        for (int i=0; i<6000; i++) {
            labels >> elm;
            if (i <num_train_sample)
                train_l.at<int>(0,i) = elm;
        }
        for (int i=0; i<6000; i++) {
            labels >> elm;
            if (i < num_test_sample)
                test_l.at<int>(0,i) = elm;
        }
    
    labels.close();
}